from fqllang.dsl.utils import getClassName
from fqllang.modules.fqlAstTranspiler.transpiler import FqlAstSqlTranspiler
from fqllang.modules.postgreSqlCodeGenerator.create import CreatePostgreSql
from fqllang.modules.postgreSqlCodeGenerator.select import SelectPostgreSql, SelectWithCriteriaPostgreSql
from fqllang.modules.sqlCodeGenerator.elements import Column, ColumnList, Condition, Criteria, DataConstraint, DataType, Field, FieldList, convertFqlDataTypeToPostgreSqlDataType

class FqlAstPostgreSqlTranspiler(FqlAstSqlTranspiler):
    def __init__(self, fqlAstModel) -> None:
        super().__init__(fqlAstModel)

    def generate(self):
        return self.fqlStatementTranspiler(self.model)

    def __str__(self) -> str:
        return self.generate()

    def fqlStatementTranspiler(self, model):
        className = getClassName(model)
        if className == "CreateForm":
            return self.createFormTranspiler(model).generate()
        elif className == "CreateNew":
            return self.createNewTranspiler(model).generate()
        elif className == "ShowForms":
            return self.showFormsTranspiler(model).generate()
        elif className == "GetCase":
            return self.getCaseTranspiler(model).generate()
        raise Exception("Modelo incorrecto")

    def createFormTranspiler(self, createForm):
        formName = createForm.formName

        dataSpecList = map(lambda dataSpec: self.dataSpecTranspiler(dataSpec), createForm.dataSpecList)

        columnList = ColumnList.emptyColumnList()

        keyField = Field('FQL_ID', DataType.serial, DataConstraint.primaryKey)
        fieldList = FieldList(keyField, *list(dataSpecList))

        return CreatePostgreSql(formName, columnList, fieldList)

    def createNewTranspiler(self, createNew):
        pass

    def getCaseTranspiler(self, getCase):
        formName=getCase.formName
        conditions = self.extractConditions(getCase)
        if conditions:
            criteria = self.getCaseConditionsTranspiler(conditions)
            return SelectWithCriteriaPostgreSql(formName, ColumnList.emptyColumnList(), criteria)
        else:
            return SelectPostgreSql(formName)
        # extendedLabelsList = self.extractExtendedLabelsList(getCase)

    def getCaseConditionsTranspiler(self, conditions):
        conditionList = []
        for condition in conditions:
            className = getClassName(condition)
            if className == "ConditionTerminal":
                attribute = condition.attribute
                valueTerminal = condition.valueTerminal
                conditionList.append(Condition(attribute, valueTerminal))
        return Criteria(*conditionList)

    def extractConditions(self, getCase):
        try:
            return getCase.conditions
        except:
            return []

    # def extractExtendedLabelsList(self, getCase):
    #     try:
    #         return getCase.extendedLabelsList
    #     except:
    #         return []

    def showFormsTranspiler(self, showForms):
        pass

    def dataSpecTranspiler(self,dataSpec):
        className = getClassName(dataSpec)
        if className == 'DataDefinition':
            return self.dataDefinitionTranspiler(dataSpec)
        elif className == 'DataReference':
            return self.dataReferenceTranspiler(dataSpec)

    def dataDefinitionTranspiler(self, dataDefinition):
        dataName = dataDefinition.dataName
        dataType = convertFqlDataTypeToPostgreSqlDataType(dataDefinition.dataType)
        notNull = dataDefinition.notNull
        unique = dataDefinition.unique
        field = Field(dataName, dataType)
        if notNull: field.addDataConstraint(DataType.notNull)
        if unique: field.addDataConstraint(DataType.unique)
        return field

    def dataReferenceTranspiler(self, dataReference):
        pass



from fqllang.modules.postgreSqlCodeGenerator.select import SelectPostgreSql, SelectWithCriteriaPostgreSql
from fqllang.modules.sqlCodeGenerator.elements import Column, ColumnList, Condition, Criteria, Where

class TestSelectPostgreSql:
    def test_generate(self):
        tableName = "Person"
        columnList = ColumnList.createColumnListByStrElements('name', 'age')
        select = SelectPostgreSql(tableName, columnList)
        print(select)
        assert select.generate() == "SELECT (name,age) FROM Person;"

    def test_generateEmpty(self):
        tableName = "Person"
        columnList = ColumnList.emptyColumnList()
        select = SelectPostgreSql(tableName, columnList)
        print(select)
        assert select.generate() == "SELECT * FROM Person;"


class TestSelectWithCriteriaPostgreSql:
    def test_generateWithEmptyColumnList(self):
        tableName = "Person"
        condition1 = Condition('name', 'Leandro')
        condition2 = Condition('age', 20)
        criteria = Criteria(condition1, condition2)
        columns = ColumnList.emptyColumnList()
        select = SelectWithCriteriaPostgreSql(tableName, columns, criteria)
        assert select.generate() == "SELECT * FROM Person WHERE name='Leandro' and age=20;"

    def test_generateWithColumnList(self):
        tableName = "Person"
        condition1 = Condition('name', 'Leandro')
        condition2 = Condition('age', 20)
        criteria = Criteria(condition1, condition2)
        columns = ColumnList.createColumnListByStrElements('name', 'firstName')
        select = SelectWithCriteriaPostgreSql(tableName, columns, criteria)
        assert select.generate() == "SELECT (name,firstName) FROM Person WHERE name='Leandro' and age=20;"

        



from fqllang.modules.sqlCodeGenerator.base import InsertSql, PreparedInsertSql
from fqllang.modules.sqlCodeGenerator.elements import Field, ColumnList


class InsertPostgreSql(InsertSql):
    def __init__(self, tableName:str, columns:ColumnList, fields:list[Field]) -> None:
        super().__init__(tableName, columns, fields)

    def _valuesList(self, fields:list[Field], columns:ColumnList):
        pass


class PreparedInsertPostgreSql(PreparedInsertSql):
    def __init__(self, tableName: str, columns: ColumnList) -> None:
        super().__init__(tableName, columns)

    def _placeholderList(self, columns:ColumnList):
        pass



import os

def getGrammar():
    with open(f"fqllang/dsl/fqlgrammar.tx") as f:
        return f.read()

def getGrammarPath():
    return "fqllang/dsl/fqlgrammar.tx"

def getClassName(object):
    return object.__class__.__name__





from fqllang.modules.sqlCodeGenerator.base import FindByKeySql, SelectSql, SelectWithCriteriaSql, SelectWithMatchSql
from fqllang.modules.sqlCodeGenerator.elements import Where


class SelectPostgreSql(SelectSql):
    def __init__(self, tableName:str, columns=None) -> None:
        super().__init__(tableName, columns)

    def generate(self):
        return f"SELECT {self.columns} FROM {self.tableName};"

    def __str__(self) -> str:
        return self.generate()

class SelectWithCriteriaPostgreSql(SelectWithCriteriaSql):
    def __init__(self, tableName: str, columns, criteria) -> None:
        super().__init__(tableName, columns, criteria)
        self.where = Where(criteria)

    def generate(self):
        return f"SELECT {self.columns} FROM {self.tableName} {self.where};"

    def __str__(self) -> str:
        return self.generate()

class SelectWithMatchPostgreSql(SelectWithMatchSql):
    def __init__(self, tableName: str, columns, pattern: str) -> None:
        super().__init__(tableName, columns, pattern)

class FindByKeyPostgreSql(FindByKeySql):
    def __init__(self,tableName:str,columns,keyColumn:str,keyValue:str):
        super().__init__(tableName, columns, keyColumn, keyValue)



from fqllang.modules.sqlCodeGenerator.elements import Column, ColumnList, Condition, Criteria, DataConstraint, DataType, Field, FieldList, Where

class TestColumnList:
    def test_len(self):
        name = Column('name')
        age = Column('age')
        columnList = ColumnList(name, age)
        assert len(columnList) == 2

    def test_isEmpty(self):
        columnList = ColumnList()
        assert columnList.isEmpty

    def test_transformColumnsInStrList(self):
        name = Column('name')
        age = Column('age')
        columnList = ColumnList(name, age)
        assert columnList.columnsInStrList() == ['name', 'age']

    def test_iter(self):
        name = Column('name')
        age = Column('age')
        columnList = ColumnList(name, age)
        for column, element in zip(columnList, [name,age]):
            assert column == element

    def test_generate(self):
        name = Column('name')
        age = Column('age')
        columnList = ColumnList(name, age)
        assert columnList.generate() == "(name,age)"

    def test_generateEmpty(self):
        columnList = ColumnList()
        assert columnList.generate() == "*"

    def test_createColumnListByStrElements(self):
        strElements = ["name", "age", "address"]
        columnList = ColumnList.createColumnListByStrElements(*strElements)
        assert columnList.generate() == "(name,age,address)"

class TestField:
    def test_generate(self):
        name = "name"
        dataType = DataType.varchar
        dataConstraint1 = DataConstraint.notNull
        dataConstraint2 = DataConstraint.unique
        field = Field(name, dataType, dataConstraint1, dataConstraint2)
        assert field.generate() == "name VARCHAR(511) NOT NULL UNIQUE"

    def test_generateWithoutDataConstraints(self):
        name = "name"
        dataType = DataType.varchar
        field = Field(name, dataType)
        assert field.generate() == "name VARCHAR(511)"

    def test_addDataConstraint(self):
        name = "name"
        dataType = DataType.varchar
        field = Field(name, dataType)
        field.addDataConstraint(DataConstraint.primaryKey)
        field.addDataConstraint(DataConstraint.primaryKey)
        assert field.generate() == "name VARCHAR(511) PRIMARY KEY"

class TestCondition:
    def test_generate(self):
        condition1 = Condition('name', 'Leandro')
        condition2 = Condition('age', 20)
        assert condition1.generate() == "name='Leandro'"
        assert condition2.generate() == "age=20"

class TestCriteria:
    def test_generate(self):
        condition1 = Condition('name', 'Leandro')
        condition2 = Condition('age', 20)
        criteria = Criteria(condition1, condition2)
        assert criteria.generate() == "name='Leandro' and age=20"

class TestWhere:
    def test_generate(self):
        condition1 = Condition('name', 'Leandro')
        condition2 = Condition('age', 20)
        criteria = Criteria(condition1, condition2)
        where = Where(criteria)
        assert where.generate() == "WHERE name='Leandro' and age=20"


class TestFieldList:
    def test_generate(self):
        field1 = Field("name", DataType.varchar, DataConstraint.notNull)
        field2 = Field("age", DataType.integer)
        field3 = Field("isCold", DataType.boolean)
        fieldList = FieldList(field1, field2, field3)
        assert fieldList.generate() == "name VARCHAR(511) NOT NULL,age INTEGER,isCold BOOLEAN"

    def test_fieldsInStrList(self):
        field1 = Field("name", DataType.varchar, DataConstraint.notNull)
        field2 = Field("age", DataType.integer)
        field3 = Field("isCold", DataType.boolean)
        fieldList = FieldList(field1, field2, field3)
        assert fieldList.fieldsInStrList() == ["name VARCHAR(511) NOT NULL", "age INTEGER", "isCold BOOLEAN"]

    def test_isEmpty(self):
        fieldList = FieldList()
        assert fieldList.isEmpty

    def test_len(self):
        field1 = Field("name", DataType.varchar, DataConstraint.notNull)
        field2 = Field("age", DataType.integer)
        field3 = Field("isCold", DataType.boolean)
        fieldList = FieldList(field1, field2, field3)
        assert len(fieldList) == 3



from fqllang.modules.sqlCodeGenerator.elements import Field, Criteria, ColumnList

class Sql:
    def __init__(self, tableName:str, columns:ColumnList) -> None:
        self.tableName = tableName
        if columns:
            self.columns = columns
        else:
            self.columns = ColumnList.emptyColumnList()

    def generate(self):
        NotImplemented()

class CreateSql(Sql):
    def __init__(self, tableName:str, columns:ColumnList) -> None:
        super().__init__(tableName, columns)

class SelectSql(Sql):
    def __init__(self, tableName:str, columns:ColumnList) -> None:
        super().__init__(tableName, columns)

class InsertSql(Sql):
    def __init__(self, tableName:str, columns:ColumnList, fields:list[Field]) -> None:
        super().__init__(tableName, columns)

    def _valuesList(self, fields:list[Field], columns:ColumnList):
        pass

class SelectWithCriteriaSql(Sql):
    def __init__(self, tableName: str, columns: ColumnList, criteria: Criteria) -> None:
        super().__init__(tableName, columns)
        self.criteria = criteria

class SelectWithMatchSql(Sql):
    def __init__(self, tableName: str, columns: ColumnList, pattern: str) -> None:
        super().__init__(tableName, columns)

class FindByKeySql(Sql):
    def __init__(self,tableName:str,columns:ColumnList,keyColumn:str,keyValue:str):
        super().__init__(tableName, columns)

class PreparedInsertSql(Sql):
    def __init__(self, tableName: str, columns: ColumnList) -> None:
        super().__init__(tableName, columns)

    def _placeholderList(self, columns:ColumnList):
        pass




from textx import metamodel_from_str, metamodel_from_file
from fqllang.dsl.utils import getGrammar, getGrammarPath

class AstBuilder:
    def __init__(self, grammar, code) -> None:
        assert grammar is not None
        assert code is not None
        self.grammar = grammar
        self.code = code

    def buildAst(self):
        NotImplemented()

class FqlAstBuilder(AstBuilder):
    def __init__(self, grammar, code) -> None:
        super().__init__(grammar, code)
        self.buildAst()

    @staticmethod
    def fqlAstFromCode(code):
        grammarPath = getGrammarPath()
        # grammar = getGrammar() # Add dependency injection
        return FqlAstBuilder(grammarPath, code)

    def buildAst(self):
        metamodel = metamodel_from_file(self.grammar)
        self.fqlAst = metamodel.model_from_str(self.code)

    @property
    def model(self):
        return self.fqlAst
